# *CoDi*: Profiling functional disruption in biological systems

**Motivation.** Biological systems comprise a vast amount of intricate processes that are deeply interconnected and carefully coordinated.
These systems must adapt to stimuli and challenge – often through dynamic functional changes in how the corresponding processes
interact. However, the associated complex disruption patterns are hard to capture, particularly when analyzing large amounts of measured
biomarkers and focusing on individual markers without accounting for their dependencies.

**Results.** To address this, we provide CoDi , an easy to use and highly customizable open-source package that allows to capture disruption
properties of biomarker networks across conditions and timepoints, and discover modules with exceptional disruption profiles. CoDi
employs the notion of correlation disruption which quantifies functional disruptions and enables a novel perspective on the coordination
and adaption of biological systems to stimuli and challenge. Our examples illustrate the scope of CoDi , by revealing coordinated functional
adaptions of the immune system during pregnancy

## Quickstart

### Install

```bash
# will be replaced by installing from PyPi upon publication
conda env create -f environment.yml --name nalab-cordis-publication
conda activate nalab-cordis-publication
python setup.py develop
```

You can now run `jupyter lab` (by default a browser should open at http://localhost:8080) and play with the notebooks in the `notebooks` folder. For example, you can find this README as a notebook: `notebooks/README.ipynb`.

If you already have a Jupyter environment set up, you can 
* register the current environment with Jupyter using the following command
* change the kernel of the notebooks accordingly within the corresponding Jupyter notebook.

```bash
# optionally register the current environment as a jupyter kernel
python -m ipykernel install --user --name=nalab-cordis-publication
```

### Load example data


```python
random_state = 43
```


```python
from nalabcordis.data.synthetic import load_example
from nalabcordis.visualization import visualize_data
```


```python
# load data
X, y, entities = load_example(random_state=random_state)

# visualize
fig, axes, correlation_matrices, coordinates = visualize_data(
    X, y, entities, mode="network", random_state=random_state);
```


    
![png](README_files/README_14_0.png)
    


### Configure and fit correlation disruption


```python
from nalabcordis.cordis.default \
    import DefaultCorrelationDisruptionFeatureTransformer \
    as CorrelationDisruptionTransformer
```


```python
# configure and fir correlation disruption transformer
disruption_transformer = CorrelationDisruptionTransformer(
    default_derive_features_aggregation="mean")
disruption_transformer.fit(X, y, groups=entities, subset_masks="y");
```


```python
# transform samples
disruption_values = disruption_transformer.transform(X)
```


```python
# we get one aggregated disruption value per cluster for each sample
disruption_values.shape
```




    (500, 3)



### Visualize disrupted modules


```python
import numpy as np
import scipy.stats

import matplotlib.pyplot as plt
import seaborn as sns

from nalabcordis.visualization import plot_cordis_cluster as plot_cluster
```


```python
y_unique = np.unique(y)

for i_cluster, cluster in enumerate(disruption_transformer.selected_clusters_):
    
    values = disruption_values[:, i_cluster]
    r, p = scipy.stats.spearmanr(values, y)
    
    fig, axes = plt.subplots(
        1, len(y_unique) + 1, 
        figsize=(4 * 1 * (len(y_unique) + 1), 4 * 1), 
        dpi=300)
 
    # correlation disruption plot
    ax = axes[0]
    x_rank = scipy.stats.rankdata(y, method="dense")
    sns.lineplot(x=x_rank, y=values, hue=entities, color="blue", alpha=0.1, ax=ax)
    sns.lineplot(x=x_rank, y=values, ax=ax)
    ax.set(
        xlabel="timepoints",
        xticks=np.unique(x_rank),
        xticklabels= y_unique,
        ylabel="disruption",
        title=f"r={r:.02f}"
    )
    ax.get_legend().remove()
    ax.spines['right'].set_visible(False)
    ax.spines['top'].set_visible(False)
    
    # cluster visualization
    for i, yy in enumerate(y_unique):
        ax = axes[i + 1]
        ax.axis("off")
        plot_cluster(
            cluster, 
            coordinates, 
            correlation_matrices[yy], 
            correlation_threshold=0, 
            verbose=0, 
            ax=ax)
    fig.suptitle(
        f"Cluster: Reference data={cluster['reference_label']}, Id={cluster['id']}")
    
    plt.show()
```


    
![png](README_files/README_22_0.png)
    



    
![png](README_files/README_22_1.png)
    



    
![png](README_files/README_22_2.png)
    


## Reproducing results and figures from the paper

### Prepare data

#### Healthy pregnancies

* For the healthy pregnancy dataset, download the data archive from: https://nalab.stanford.edu/multiomics-pregnancy/
* Extract the archive and put the `Data.Rda` file into the `data/multiomics` folder.

#### Preeclamptic pregnancies

* For the preeclampsia dataset, download the data archive from: https://nalab.stanford.edu/codi/
* Extract the archive and put the `2019-12-04_Preeclampsia-data.rda` file into the `data/preeclampsia` folder.

### Experiments on real-world data

**Notebooks:**
* `notebooks/01_exp_regression.ipynb`
* `notebooks/02_exp_classification.ipynb`

The code for reproducing the results on the real world datasets are found in the above mentioned notebooks.

Please start Jupyter as outlined in [Section 1.1.1](#Install) and run the notebooks, or use `nbconvert --execute` or `papermill`.
The produced figures will be in `_out/realworld`.

### Experiments on synthetic data

**Notebooks:**
* `notebooks/03_synthetic.ipynb`

The code for reproducing the results on synthetic data that show the properties and characteristics of the proposed correlation disruption method are found in the above mentioned notebook.

Please start Jupyter as outlined in [Section 1.1.1](#Install) and run the notebook, or use `nbconvert --execute` or `papermill`.
The produced figures will be in `_out/synthetic`.

### Figure 1

**Notebooks:**
* `notebooks/04_figure1.ipynb`

The code for reproducing the panels in Figure 1 from the manuscript are found in the above mentioned notebook.

Please start Jupyter as outlined in [Section 1.1.1](#Install) and run the notebook, or use `nbconvert --execute` or `papermill`.
The produced figures will be in `_out/figure1`.
